__title__ = "MindsDB Lindorm handler"
__package_name__ = "mindsdb_lindorm_handler"
__version__ = "0.0.1"
__description__ = "MindsDB handler for Lindorm"
__author__ = "Someshfengde"
__github__ = "https://github.com/someshfengde"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "MIT"
__copyright__ = "Copyright 2023 - mindsdb"
