__title__ = 'MindsDB CoinBase handler'
__package_name__ = 'mindsdb_coinbase_handler'
__version__ = '0.0.1'
__description__ = "MindsDB handler for the CoinBase API"
__author__ = 'Bustedware LLC'
__github__ = 'https://github.com/bustedware'
__pypi__ = 'https://pypi.org/project/mindsdb/'
__license__ = 'MIT'
__copyright__ = 'Copyright 2023 - mindsdb'
